@@ -1,4 +1,4 @@
-/* 
+/*
  * Copyright (C) 2006, 2007 OpenedHand Ltd.
  *
  * Author: Jorn Baayen <jorn@openedhand.com>
@@ -31,21 +31,26 @@
 #include <errno.h>
 #include <string.h>
 #include <unistd.h>
+#include <fcntl.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <sys/utsname.h>
-#include <sys/ioctl.h> 
+#include <sys/ioctl.h>
 #include <sys/socket.h>
 #include <sys/types.h>
+#include <sys/stat.h>
 #include <net/if.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <libsoup/soup-address.h>
 
+#include "xdgmime/xdgmime.h"
+
 #include "gupnp-context.h"
 #include "gupnp-context-private.h"
 #include "gupnp-marshal.h"
 #include "gena-protocol.h"
+#include "accept-language.h"
 
 G_DEFINE_TYPE (GUPnPContext,
                gupnp_context,
@@ -86,7 +91,7 @@ make_server_id (void)
         struct utsname sysinfo;
 
         uname (&sysinfo);
-        
+
         return g_strdup_printf ("%s/%s UPnP/1.0 GUPnP/%s",
                                 sysinfo.sysname,
                                 sysinfo.version,
@@ -101,29 +106,29 @@ make_server_id (void)
 gboolean
 get_ip (const char *iface, ip_address ip)
 {
-	struct ifreq *ifr;
-	struct ifreq ifrr;
-	struct sockaddr_in sa;
-	struct sockaddr ifaddr;
-	int sockfd;
+        struct ifreq *ifr;
+        struct ifreq ifrr;
+        struct sockaddr_in sa;
+        struct sockaddr ifaddr;
+        int sockfd;
 
-	if ((sockfd = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
-	        return FALSE;
+        if ((sockfd = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
+                return FALSE;
 
-	ifr = &ifrr;
+        ifr = &ifrr;
 
-	ifrr.ifr_addr.sa_family = AF_INET;
+        ifrr.ifr_addr.sa_family = AF_INET;
 
-	strncpy (ifrr.ifr_name, iface, sizeof (ifrr.ifr_name));
+        strncpy (ifrr.ifr_name, iface, sizeof (ifrr.ifr_name));
 
-	if (ioctl (sockfd, SIOCGIFADDR, ifr) < 0)
-		return FALSE;
+        if (ioctl (sockfd, SIOCGIFADDR, ifr) < 0)
+                return FALSE;
 
-	ifaddr = ifrr.ifr_addr;
-	strncpy (ip, inet_ntoa (inaddrr (ifr_addr.sa_data)),
+        ifaddr = ifrr.ifr_addr;
+        strncpy (ip, inet_ntoa (inaddrr (ifr_addr.sa_data)),
                  sizeof (ip_address));
 
-	return TRUE;
+        return TRUE;
 }
 
 /**
@@ -135,7 +140,7 @@ get_default_host_ip (void)
         struct if_nameindex *ifs;
         char *ret;
         int i;
-        
+
         /* List network interfaces */
         ifs = if_nameindex ();
         if (ifs == NULL) {
@@ -165,6 +170,11 @@ get_default_host_ip (void)
 
         if_freenameindex (ifs);
 
+        if (!ret) {
+                /* Didn't find anything. Let's take the loopback IP. */
+                ret = g_strdup (LOOPBACK_IP);
+        }
+
         return ret;
 }
 
@@ -178,16 +188,8 @@ gupnp_context_init (GUPnPContext *context)
                                              GUPNP_TYPE_CONTEXT,
                                              GUPnPContextPrivate);
 
-        context->priv->host_ip = NULL;
-        context->priv->port    = SOUP_ADDRESS_ANY_PORT;
-
-        context->priv->subscription_timeout = GENA_DEFAULT_TIMEOUT;
-
         context->priv->session = soup_session_async_new ();
 
-        context->priv->server     = NULL; /* Started on demand */
-        context->priv->server_url = NULL;
-
         server_id = make_server_id ();
         gssdp_client_set_server_id (GSSDP_CLIENT (context), server_id);
         g_free (server_id);
@@ -253,6 +255,7 @@ static void
 gupnp_context_dispose (GObject *object)
 {
         GUPnPContext *context;
+        GObjectClass *object_class;
 
         context = GUPNP_CONTEXT (object);
 
@@ -265,18 +268,27 @@ gupnp_context_dispose (GObject *object)
                 g_object_unref (context->priv->server);
                 context->priv->server = NULL;
         }
+
+        /* Call super */
+        object_class = G_OBJECT_CLASS (gupnp_context_parent_class);
+        object_class->dispose (object);
 }
 
 static void
 gupnp_context_finalize (GObject *object)
 {
         GUPnPContext *context;
+        GObjectClass *object_class;
 
         context = GUPNP_CONTEXT (object);
 
         g_free (context->priv->host_ip);
 
         g_free (context->priv->server_url);
+
+        /* Call super */
+        object_class = G_OBJECT_CLASS (gupnp_context_parent_class);
+        object_class->finalize (object);
 }
 
 static void
@@ -306,7 +318,10 @@ gupnp_context_class_init (GUPnPContextClass *klass)
                                       "The local host's IP address",
                                       NULL,
                                       G_PARAM_READWRITE |
-                                      G_PARAM_CONSTRUCT_ONLY));
+                                      G_PARAM_CONSTRUCT_ONLY |
+                                      G_PARAM_STATIC_NAME |
+                                      G_PARAM_STATIC_NICK |
+                                      G_PARAM_STATIC_BLURB));
 
         /**
          * GUPnPContext:port
@@ -321,7 +336,10 @@ gupnp_context_class_init (GUPnPContextClass *klass)
                                     "Port to run on",
                                     0, G_MAXUINT, SOUP_ADDRESS_ANY_PORT,
                                     G_PARAM_READWRITE |
-                                    G_PARAM_CONSTRUCT_ONLY));
+                                    G_PARAM_CONSTRUCT_ONLY |
+                                    G_PARAM_STATIC_NAME |
+                                    G_PARAM_STATIC_NICK |
+                                    G_PARAM_STATIC_BLURB));
 
         /**
          * GUPnPContext:subscription-timeout
@@ -335,9 +353,14 @@ gupnp_context_class_init (GUPnPContextClass *klass)
                  g_param_spec_uint ("subscription-timeout",
                                     "Subscription timeout",
                                     "Subscription timeout",
-                                    0, G_MAXUINT, GENA_DEFAULT_TIMEOUT,
+                                    GENA_MIN_TIMEOUT,
+                                    G_MAXUINT,
+                                    GENA_DEFAULT_TIMEOUT,
                                     G_PARAM_READWRITE |
-                                    G_PARAM_CONSTRUCT_ONLY));
+                                    G_PARAM_CONSTRUCT_ONLY |
+                                    G_PARAM_STATIC_NAME |
+                                    G_PARAM_STATIC_NICK |
+                                    G_PARAM_STATIC_BLURB));
 }
 
 SoupSession *
@@ -347,14 +370,14 @@ _gupnp_context_get_session (GUPnPContext *context)
 }
 
 /**
- * Default server handler: Return 501 not implemented.
+ * Default server handler: Return 404 not found.
  **/
 static void
 default_server_handler (SoupServerContext *server_context,
                         SoupMessage       *msg,
                         gpointer           user_data)
 {
-        soup_message_set_status (msg, SOUP_STATUS_NOT_IMPLEMENTED);
+        soup_message_set_status (msg, SOUP_STATUS_NOT_FOUND);
 }
 
 SoupServer *
@@ -370,7 +393,7 @@ _gupnp_context_get_server (GUPnPContext *context)
 
                 soup_server_run_async (context->priv->server);
         }
-        
+
         return context->priv->server;
 }
 
@@ -495,3 +518,212 @@ gupnp_context_get_subscription_timeout (GUPnPContext *context)
 
         return context->priv->subscription_timeout;
 }
+
+/* Modified from libsoup simple-httpd.c */
+static void
+hosting_server_handler (SoupServerContext *server_context,
+                        SoupMessage       *msg,
+                        gpointer           user_data)
+{
+        const char *local_path, *lang, *mime;
+        char *path, *path_to_open, *path_locale, *slash;
+        SoupMethodId method;
+        struct stat st;
+        int fd, path_offset;
+        GList *locales;
+
+        local_path = (const char *) user_data;
+
+        path = soup_uri_to_string (soup_message_get_uri (msg), TRUE);
+
+        method = soup_method_get_id (msg->method);
+        if (method != SOUP_METHOD_ID_GET && method != SOUP_METHOD_ID_HEAD) {
+                soup_message_set_status (msg, SOUP_STATUS_NOT_IMPLEMENTED);
+                locales = NULL;
+                goto DONE;
+        }
+
+        /* Get preferred locales */
+        locales = accept_language_get_locales (msg);
+
+        if (path) {
+                if (*path != '/') {
+                        soup_message_set_status (msg, SOUP_STATUS_BAD_REQUEST);
+                        goto DONE;
+                }
+
+                /* Skip the server path */
+                path_offset = strlen (server_context->handler->path);
+        } else {
+                path = g_strdup ("");
+
+                path_offset = 0;
+        }
+
+        path_locale = NULL;
+
+ NEXT_LOCALE:
+        g_free (path_locale);
+
+        if (locales) {
+                path_locale = g_strdup_printf ("%s.%s",
+                                               path + path_offset,
+                                               (char *) locales->data);
+        } else
+                path_locale = g_strdup (path + path_offset);
+
+        path_to_open = g_build_filename (local_path, path_locale, NULL);
+
+ AGAIN:
+        if (stat (path_to_open, &st) == -1) {
+                g_free (path_to_open);
+                if (errno == EPERM)
+                        soup_message_set_status (msg, SOUP_STATUS_FORBIDDEN);
+                else if (errno == ENOENT) {
+                        if (locales) {
+                                locales = locales->next;
+                                goto NEXT_LOCALE;
+                        } else
+                                soup_message_set_status (msg,
+                                                         SOUP_STATUS_NOT_FOUND);
+                } else
+                        soup_message_set_status
+                                (msg, SOUP_STATUS_INTERNAL_SERVER_ERROR);
+                goto DONE;
+        }
+
+        /* Handle directories */
+        if (S_ISDIR (st.st_mode)) {
+                slash = strrchr (path_to_open, '/');
+                if (!slash || slash[1]) {
+                        char *uri, *redir_uri;
+
+                        uri = soup_uri_to_string (soup_message_get_uri (msg),
+                                                  FALSE);
+                        redir_uri = g_strdup_printf ("%s/", uri);
+                        soup_message_add_header (msg->response_headers,
+                                                 "Location", redir_uri);
+                        soup_message_set_status (msg,
+                                                 SOUP_STATUS_MOVED_PERMANENTLY);
+                        g_free (redir_uri);
+                        g_free (uri);
+                        g_free (path_to_open);
+                        goto DONE;
+                }
+
+                g_free (path_to_open);
+                path_to_open = g_build_filename (local_path,
+                                                 path_locale,
+                                                 "index.html", NULL);
+                goto AGAIN;
+        }
+
+        g_free (path_locale);
+
+        /* Read file */
+        fd = open (path_to_open, O_RDONLY);
+        g_free (path_to_open);
+        if (fd == -1) {
+                soup_message_set_status (msg,
+                                         SOUP_STATUS_INTERNAL_SERVER_ERROR);
+                goto DONE;
+        }
+
+        msg->response.owner = SOUP_BUFFER_SYSTEM_OWNED;
+        msg->response.length = st.st_size;
+
+        msg->response.body = g_malloc (msg->response.length);
+        read (fd, msg->response.body, msg->response.length);
+
+        close (fd);
+
+        /* Set Content-Type */
+        mime = xdg_mime_get_mime_type_for_data (msg->response.body,
+                                                msg->response.length);
+
+        soup_message_add_header (msg->response_headers,
+                                 "Content-Type",
+                                 mime);
+
+        /* Set Content-Language */
+        if (locales) {
+                http_language_from_locale (locales->data);
+
+                lang = locales->data;
+        } else
+                lang = "en";
+
+        soup_message_add_header (msg->response_headers,
+                                 "Content-Language",
+                                 lang);
+
+        /* OK */
+        soup_message_set_status (msg, SOUP_STATUS_OK);
+
+ DONE:
+        /* Cleanup */
+        g_free (path);
+
+        while (locales) {
+                g_free (locales->data);
+                locales = g_list_delete_link (locales, locales);
+        }
+}
+
+static void
+unregister_hosting_server_handler (SoupServer        *server,
+                                   SoupServerHandler *handler,
+                                   gpointer           user_data)
+{
+        g_free (user_data);
+}
+
+/**
+ * gupnp_context_host_path
+ * @context: A #GUPnPContext
+ * @local_path: Path to the local file or folder to be hosted
+ * @server_path: Web server path where @local_path should be hosted
+ *
+ * Start hosting @local_path at @server_path. Files with the path
+ * @local_path.LOCALE (if they exist) will be served up when LOCALE is
+ * specified in the request's Accept-Language header.
+ **/
+void
+gupnp_context_host_path (GUPnPContext *context,
+                         const char   *local_path,
+                         const char   *server_path)
+{
+        SoupServer *server;
+
+        g_return_if_fail (GUPNP_IS_CONTEXT (context));
+        g_return_if_fail (local_path != NULL);
+        g_return_if_fail (server_path != NULL);
+
+        server = _gupnp_context_get_server (context);
+
+        soup_server_add_handler (server, server_path, NULL,
+                                 hosting_server_handler,
+                                 unregister_hosting_server_handler,
+                                 g_strdup (local_path));
+}
+
+/**
+ * gupnp_context_unhost_path
+ * @context: A #GUPnPContext
+ * @server_path: Web server path where the file or folder is hosted
+ *
+ * Stop hosting the file or folder at @server_path.
+ **/
+void
+gupnp_context_unhost_path (GUPnPContext *context,
+                           const char   *server_path)
+{
+        SoupServer *server;
+
+        g_return_if_fail (GUPNP_IS_CONTEXT (context));
+        g_return_if_fail (server_path != NULL);
+
+        server = _gupnp_context_get_server (context);
+
+        soup_server_remove_handler (server, server_path);
+}
